"""Fondat module for Amazon Simple Queue Service (SQS)."""

import logging

from collections.abc import Iterable
from fondat.aws import Service
from fondat.codec import String, get_codec
from fondat.error import InternalServerError
from fondat.resource import resource, mutation
from fondat.security import Policy
from typing import Annotated, Optional


_logger = logging.getLogger(__name__)


def queue_resource(
    *,
    service: Service = None,
    queue_url: str,
    message_type: type,
    policies: Iterable[Policy] = None,
):
    """
    Create SQS queue resource.

    Parameters:
    • service: S3 service object
    • queue_url: the URL of the SQS queue
    • message_type: type of value transmitted in each message
    • security: security policies to apply to all operations
    """

    if service is None:
        service = Service("sqs")

    if service.name != "sqs":
        raise TypeError("expecting sqs service object")

    codec = get_codec(String, message_type)

    @resource
    class Queue:
        """SQS queue resource."""

        @mutation(policies=policies)
        async def send(self, message: message_type) -> None:
            """Send a message to the queue."""
            client = await service.client()
            try:
                await client.send_message(QueueUrl=queue_url, MessageBody=codec.encode(message))
            except Exception as e:
                _logger.error(e)
                raise InternalServerError from e

        @mutation(policies=policies)
        async def receive(
            self,
            wait_time_seconds: Annotated[int, "number of seconds to wait for a message"] = 0,
        ) -> Optional[message_type]:
            """Return the next message from the queue."""
            client = await service.client()
            response = await client.receive_message(
                QueueUrl=queue_url, WaitTimeSeconds=wait_time_seconds
            )
            if "Messages" not in response:
                return None
            try:
                result = codec.decode(response["Messages"][0]["Body"])
            except Exception as e:
                _logger.error(e)
                raise InternalServerError from e
            return result

    return Queue()



"""Fondat module for Amazon Simple Email Service (SES)."""

import logging

from collections.abc import Iterable
from fondat.aws import Service, wrap_client_error
from fondat.http import AsBody
from fondat.resource import operation, resource, mutation
from fondat.security import Policy
from typing import Annotated, Union


_logger = logging.getLogger(__name__)


def ses_resource(
    service: Service = None,
    policies: Iterable[Policy] = None,
):
    """
    Create SES resource.

    Parameters:
    • service: SES service object
    • policies: security policies to apply to all operations
    """

    if service is None:
        service = Service("ses")

    if service.name != "ses":
        raise TypeError("expecting ses service object")

    @resource
    class Identity:
        """Verified identity."""

        def __init__(self, identity: str):
            self.identity = identity

        @operation(policies=policies)
        async def delete(self):
            """Delete the identity from verified identities."""
            client = await service.client()
            await client.delete_identity(Identity=self.identity)

    @resource
    class Identities:
        """Identities for Amazon SES account in a specific region."""

        @operation(policies=policies)
        async def post(self, identity):
            """Add email address to list of identities for SES account."""
            client = await service.client()
            await client.verify_email_identity(EmailAddress=identity)

        def __getitem__(self, identity) -> Identity:
            return Identity(identity)

    @resource
    class SESResource:
        """Simple Email Service (SES) resource."""

        @mutation(policies=policies)
        async def send_raw_email(
            self,
            source: str,
            destinations: Union[str, Iterable[str]],
            data: Annotated[bytes, AsBody],
        ):
            """
            Compose an email message and immediately queue it for sending.

            Parameters:
            • source: email address to send message message from
            • desinations: email address(es) to send message to
            • data: byte string containing headers and body of message to send
            """

            if isinstance(destinations, str):
                destinations = [destinations]

            client = await service.client()
            with wrap_client_error():
                await client.send_raw_email(
                    Source=source, Destinations=destinations, RawMessage={"Data": data}
                )

        identities = Identities()

    return SESResource()



"""Fondat module for Amazon Simple Storage Service (S3)."""

import fondat.codec
import fondat.pagination
import logging

from collections.abc import Iterable
from fondat.aws import Service
from fondat.codec import Binary, String
from fondat.error import InternalServerError, NotFoundError
from fondat.resource import resource, operation
from fondat.security import Policy
from typing import Any
from urllib.parse import quote


_logger = logging.getLogger(__name__)


def bucket_resource(
    *,
    service: Service = None,
    bucket: str,
    folder: str = None,
    key_type: type,
    value_type: type,
    extension: str = None,
    compress: Any = None,
    encode_keys: bool = False,
    policies: Iterable[Policy] = None,
):
    """
    Create S3 bucket resource.

    Parameters:
    • service: S3 service object
    • bucket: name of bucket to contain objects
    • folder: name of folder within bucket to contain objects
    • key_type: type of key to identify object
    • value_type: type of value stored in each object
    • extenson: filename extension to use for each file (including dot)
    • compress: algorithm to compress and decompress content
    • encode_keys: URL encode and decode object keys
    • security: security policies to apply to all operations
    """

    if service is None:
        service = Service("s3")

    if service.name != "s3":
        raise TypeError("expecting s3 service object")

    key_codec = fondat.codec.get_codec(String, key_type)
    value_codec = fondat.codec.get_codec(Binary, value_type)

    @resource
    class Object:
        """S3 object resource."""

        def __init__(self, key: key_type):
            key = key_codec.encode(key)
            if encode_keys:
                key = quote(key, safe="")
            if extension is not None:
                key = f"{key}{extension}"
            if folder is not None:
                key = f"{folder}/{key}"
            self.key = key

        @operation(policies=policies)
        async def get(self) -> value_type:
            client = await service.client()
            try:
                response = await client.get_object(Bucket=bucket, Key=self.key)
                async with response["Body"] as stream:
                    body = await stream.read()
                if compress:
                    body = compress.decompress(body)
                return value_codec.decode(body)
            except client.exceptions.NoSuchKey:
                raise NotFoundError
            except Exception as e:
                _logger.error(e)
                raise InternalServerError from e

        @operation(policies=policies)
        async def put(self, value: value_type) -> None:
            body = value_codec.encode(value)
            if compress:
                body = compress.compress(body)
            client = await service.client()
            try:
                await client.put_object(Bucket=bucket, Key=self.key, Body=body)
            except Exception as e:
                _logger.error(e)
                raise InternalServerError from e

        @operation(policies=policies)
        async def delete(self) -> None:
            client = await service.client()
            try:
                await client.delete_object(Bucket=bucket, Key=self.key)
            except Exception as e:
                _logger.error(e)
                raise InternalServerError from e

    def _prefix(prefix):
        if not folder:
            return prefix
        if not prefix:
            return folder
        return f"{folder}/{prefix}"

    key_offset = len(folder) + 1 if folder else 0
    Page = fondat.pagination.make_page_dataclass("Page", str)

    @resource
    class Bucket:
        """S3 bucket resource."""

        @operation(policies=policies)
        async def get(
            self, prefix: str = None, limit: int = None, cursor: bytes = None
        ) -> Page:
            kwargs = {}
            if limit and limit > 0:
                kwargs["MaxKeys"] = limit
            if prefix := _prefix(prefix):
                kwargs["Prefix"] = prefix
            if cursor is not None:
                kwargs["ContinuationToken"] = cursor.decode()
            client = await service.client()
            try:
                response = await client.list_objects_v2(Bucket=bucket, **kwargs)
                next_token = response.get("NextContinuationToken")
                page = Page(
                    items=[
                        content["Key"][key_offset:] for content in response.get("Contents", ())
                    ],
                    cursor=next_token.encode() if next_token else None,
                    remaining=None,
                )
                return page
            except Exception as e:
                _logger.error(e)
                raise InternalServerError from e

        def __getitem__(self, key: key_type) -> Object:
            return Object(key)

    return Bucket()



"""AWS Lambda module."""

import asyncio
import fondat.context
import fondat.http

from base64 import b64encode, b64decode
from collections.abc import Callable, Coroutine
from fondat.error import InternalServerError
from fondat.stream import BytesStream
from typing import Any


def async_function(
    handler: Callable[[dict[str, Any], Any], Coroutine[Any, Any, Any]],
    init: Callable[[], Coroutine[Any, Any, None]] = None,
    loop: asyncio.AbstractEventLoop = None,
):
    """
    Return an AWS Lambda function that invokes an asynchronous coroutine function with event
    and context.

    Parameters:
    • coroutine: coroutine function to invoke for each call
    • init: initialization coroutine function to invoke prior to first coroutine invocation
    • loop: event loop to run function in; None creates a new one
    """

    def function(event: dict[str, Any], context: Any) -> dict[str, Any]:
        nonlocal loop
        nonlocal init
        if not loop:
            loop = asyncio.new_event_loop()
        if init:
            loop.run_until_complete(init())
            init = None
        return loop.run_until_complete(handler(event, context))

    return function


def http_function(
    handler: Callable[[fondat.http.Request], Coroutine[Any, Any, fondat.http.Response]],
    init: Callable[[], Coroutine[Any, Any, None]] = None,
) -> Callable[[Any, Any], Coroutine[Any, Any, dict[str, Any]]]:
    """
    Return an AWS Lambda function to invoke an HTTP request handler.

    Parameters:
    • handler: HTTP request handler coroutine function to invoke for each call
    • init: coroutine function to invoke prior to first request handler invocation
    """

    async def http_handler(event: dict[str, Any], context: Any) -> dict[str, Any]:
        if event["version"] != "2.0":
            raise InternalServerError("expecting payload version: 2.0")
        with fondat.context.push(
            {
                "context": "fondat.aws.lambda.http",
                "lambda_event": event,
                "lambda_context": context,
            }
        ):
            request = fondat.http.Request()
            http = event["requestContext"]["http"]
            protocol, version = http["protocol"].split("/", 1)
            if protocol != "HTTP":
                raise InternalServerError("expecting HTTP protocol")
            request.method = http["method"]
            request.path = http["path"]
            request.version = version
            for key, value in event["headers"].items():
                request.headers[key] = value
            for cookie in event.get("cookies", ()):
                request.cookies.load(cookie)
            for key, value in event.get("queryStringParameters", {}).items():
                request.query[key] = value
            body = event.get("body")
            if body:
                request.body = BytesStream(
                    b64decode(body) if event["isBase64Encoded"] else body.encode(),
                    request.headers.get("content-length"),
                )
            response = await handler(request)
            headers = response.headers
            return {
                "isBase64Encoded": True,
                "statusCode": response.status,
                "headers": {k: ", ".join(headers.getall(k)) for k in headers.keys()},
                "body": (
                    b64encode(b"".join([b async for b in response.body])).decode()
                    if response.body is not None
                    else ""
                ),
            }

    return async_function(http_handler, init)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fondat', 'fondat.aws']

package_data = \
{'': ['*']}

install_requires = \
['aiobotocore>=2.0,<3.0', 'fondat-core>=3.0,<4.0']

setup_kwargs = {
    'name': 'fondat-aws',
    'version': '3.0b8',
    'description': 'Fondat package for Amazon Web Services.',
    'long_description': '# fondat-aws\n\n[![PyPI](https://badge.fury.io/py/fondat-aws.svg)](https://badge.fury.io/py/fondat-aws)\n[![Python](https://img.shields.io/pypi/pyversions/fondat-core)](https://python.org/)\n[![GitHub](https://img.shields.io/badge/github-main-blue.svg)](https://github.com/fondat/fondat-aws/)\n[![Test](https://github.com/fondat/fondat-aws/workflows/test/badge.svg)](https://github.com/fondat/fondat-aws/actions?query=workflow/test)\n[![License](https://img.shields.io/github/license/fondat/fondat-aws.svg)](https://github.com/fondat/fondat-aws/blob/main/LICENSE)\n[![Black](https://img.shields.io/badge/code%20style-black-black.svg)](https://github.com/psf/black)\n\nFondat package for Amazon Web Services.\n\n## Develop\n\n```\npoetry install\npoetry run pre-commit install\n```\n\n## Test\n\n```\npoetry run pytest\n```\n',
    'author': 'fondat-aws authors',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fondat/fondat-aws/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Fondat module for AWS CloudWatch."""

import logging

from collections import deque
from collections.abc import Iterable
from datetime import datetime
from fondat.aws import Service
from fondat.data import datacls
from fondat.resource import resource, operation, mutation
from fondat.security import Policy
from typing import Any, Literal, Optional, Union
from fondat.monitoring import Measurement, Counter, Gauge, Absolute


_logger = logging.getLogger(__name__)


Unit = Literal[
    "Seconds",
    "Microseconds",
    "Milliseconds",
    "Bytes",
    "Kilobytes",
    "Megabytes",
    "Gigabytes",
    "Terabytes",
    "Bits",
    "Kilobits",
    "Megabits",
    "Gigabits",
    "Terabits",
    "Percent",
    "Count",
    "Bytes/Second",
    "Kilobytes/Second",
    "Megabytes/Second",
    "Gigabytes/Second",
    "Terabytes/Second",
    "Bits/Second",
    "Kilobits/Second",
    "Megabits/Second",
    "Gigabits/Second",
    "Terabits/Second",
    "Count/Second",
]

Value = Union[int, float]

Values = dict[Union[int, float], Union[int, float]]  # value: count


@datacls
class Statistics:
    """
    Statistics measurement type.
    Attributes:
    • count: count of measured values
    • sum: sum of all measured values
    • minimum: minimum measured value
    • maximum: maximum measured value
    """

    count: Union[int, float]
    sum: Union[int, float]
    minimum: Union[int, float]
    maximum: Union[int, float]


@datacls
class Metric:
    """
    CloudWatch metric type.
    Attributes:
    • name: name of the CloudWatch metric
    • dimensions: name/value pair of categories for characteristics
    • value: value of measurement
    • timestamp: date and time of the measurement to record
    • unit: Optional[Unit]
    • resolution: granularity of the metric
    """

    name: str
    dimensions: dict[str, str]
    value: Union[Value, Values, Statistics]
    timestamp: datetime
    unit: Optional[Unit]
    resolution: Optional[int]


def cloudwatch_resource(
    *,
    service: Service = None,
    policies: Iterable[Policy] = None,
):
    """
    Create CloudWatch resource.

    Parameters:
    • service: CloudWatch service object
    • security: security requirements to apply to all operations
    """

    if service is None:
        service = Service("cloudwatch")

    if service.name != "cloudwatch":
        raise TypeError("expecting cloudwatch service object")

    @resource
    class NamespaceResource:
        """Create Namespace resource."""

        def __init__(self, name: str):
            self.name = name

        @operation(policies=policies)
        async def post(self, metrics: Iterable[Metric]):
            metrics = deque(metrics)
            data = []
            while metrics:
                metric = metrics.popleft()
                datum = {
                    "MetricName": metric.name,
                    "Dimensions": [
                        {"Name": k, "Value": v} for k, v in metric.dimensions.items()
                    ],
                    "Timestamp": metric.timestamp,
                }
                if metric.unit:
                    datum["Unit"] = metric.unit
                if metric.resolution:
                    datum["Resolution"] = metric.resolution
                if isinstance(metric.value, (int, float)):
                    datum["Value"] = float(metric.value)
                elif isinstance(metric.value, dict):
                    datum["Values"] = [float(v) for v in metric.value.keys()]
                    datum["Counts"] = [float(v) for v in metric.value.values()]
                elif isinstance(metric.value, Statistics):
                    datum["StatisticValues"] = {
                        "SampleCount": float(metric.value.count),
                        "Sum": float(metric.value.sum),
                        "Minimum": float(metric.value.minimum),
                        "Maximum": float(metric.value.maximum),
                    }
                data.append(datum)
                if len(data) == 20 or not metrics:
                    client = await service.client()
                    await client.put_metric_data(Namespace=self.name, MetricData=data)

    @resource
    class CloudWatchResource:
        """Create CloudWatch resource."""

        def namespace(self, name: str) -> NamespaceResource:
            return NamespaceResource(name)

    return CloudWatchResource()


class CloudWatchMonitor:
    """
    A monitor that stores all recorded measurements in CloudWatch.
    """

    # future: collect metrics, send in batches

    def __init__(self, service: Service, namespace: str):
        self.resource = cloudwatch_resource(service=service).namespace(namespace)

    async def record(self, measurement: Measurement):
        """Record a measurement."""

        if measurement.type == "counter":
            m = Counter(timestamp=measurement.timestamp)
            m.record(measurement.value)
            metric = Metric(
                name=measurement.tags["name"],
                dimensions={"Name": measurement.type, "Value": str(m.value)},
                timestamp=measurement.timestamp,
                value=float(m.value),
                unit="Count",
            )
        elif measurement.type == "gauge":
            m = Gauge(timestamp=measurement.timestamp)
            m.record(measurement.value)
            metric = Metric(
                name=measurement.tags["name"],
                dimensions={"Name": measurement.type, "Value": str(measurement.value)},
                timestamp=measurement.timestamp,
                value=Statistics(
                    count=float(m.count),
                    sum=float(m.sum),
                    minimum=float(m.min),
                    maximum=float(m.max),
                ),
            )
        elif measurement.type == "absolute":
            m = Absolute(timestamp=measurement.timestamp)
            m.record(measurement.value)
            metric = Metric(
                name=measurement.tags["name"],
                dimensions={"Name": measurement.type, "Value": str(m.value)},
                timestamp=measurement.timestamp,
                value=float(m.value),
            )

        await self.resource.post(metrics=[metric])
